/*
 * kerndig.c - implementation of digging in the kernel to look up or
 * by Jonathan Kay, Computer Systems Lab, UCSD         8/3/93
 *
 * Copyright (c) 1993,1994 Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * paragraph appears in all copies of this software.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE REGENTS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <nlist.h>
#include <fcntl.h>
#include "error.h"
#include "vars.h"

#ifndef KMEM
#define KMEM "/dev/kmem"
#endif

extern char *cmdname;
extern char *kname;
extern int existwarn;

int kernel_init(), kernel_get(), kernel_set();

struct var *kreadq = 0, *kwriteq = 0;
int nkread = 0, nkwrite = 0;
static int kfd;

/*
 * Open /dev/kmem with whatever permissions are appropriate
 */
kernel_init()
{
    int openflags = 0;
    int uid, euid, gid, egid;

    /*
     * Figure out needed permissions flags.  We can't just open RDWR because
     * we don't want just anybody to be able to change this, yet we do want
     * people who can write to kmem (e.g., in a trusting environment) to be able
     * to change options.
     */
    if (kwriteq)
	openflags = kreadq ? O_RDWR : O_WRONLY;
    else if (kreadq)
	openflags = O_RDONLY;
    else
	return;

    /*
     * Permissions scheme.  We want to let anybody read the options.
     * However, one should not be able to change options unless one
     * is root or otherwise has permission to read /dev/kmem.
     * Since we can only implement that function through being setuid,
     * we must reduce our permissions to uid for change attempts.
     */
    if (kwriteq) {
	/* set uid */
	uid = getuid();
	euid = geteuid();
	if (euid != uid)
	    /* This program has been installed setuid. */
	    setuid(uid);

	/* set gid */
	gid = getgid();
	egid = getegid();
	if (egid != gid)
	    /* This program has been installed setgid. */
	    setgid(gid);
    }

    if ((kfd = open(KMEM, openflags)) == -1)
	SYSERR(KMEM);
}

kernel_get(vars, nnames)
Var *vars;
int nnames;
{
    Var *v;
    char **inamep;
    struct nlist *nls, *nl;
    int nbadsyms;

    nl = nls = (struct nlist *) malloc((nnames + 1) * sizeof(*nl));
    for (v = vars; v; v = v->next) {
	v->foundname = 0;
	for (inamep = v->inames; *inamep; ++inamep)
	    (nl++)->n_name = *inamep;
    }
    nl->n_name = "";
    if (kname) {
	if ((nbadsyms = nlist(kname, nls)) == -1)
	    goto badkern;
    }
    else {
	if ((nbadsyms = nlist("/vmunix", nls)) == -1
	    && (nbadsyms = nlist("/unix", nls)) == -1
	    && (nbadsyms = nlist("/bsd", nls)) == -1
	    && (nbadsyms = nlist("/hp-ux", nls)) == -1)
	    goto badkern;
    }
    /* XXX - restore for verbose mode */
/*    printf("couldn't get %d symbols\n", nbadsyms);*/
    nl = nls;
    for (v = vars; v; v = v->next) {
	for (inamep = v->inames; *inamep; ++inamep) {
	    if (nl->n_type == 0L) {
		/* XXX - restore for verbose mode */
/*		fprintf(stderr, "couldn't find %s\n", nl->n_name);*/
	    }
	    else {
		lseek(kfd, nl->n_value, 0);
		if (read(kfd, v->value, v->len) < v->len)
		    fprintf(stderr, "%s: couldn't read %s\n",
			    cmdname, nl->n_name);
		else
		    v->foundname = *inamep;
	    }
	    ++nl;
	}
	if (!v->foundname && existwarn)
	    fprintf(stderr, "%s: couldn't find %s to read\n",
		    cmdname, v->inames[0]);
    }
    free(nls);
    return;
  badkern:
    perror("netconfig: nlist");
    fprintf(stderr, "%s: could not get symbols from kernel\n", cmdname);
}

kernel_set(vars, nnames)
Var *vars;
int nnames;
{
    Var *v;
    char **inamep;
    struct nlist *nls, *nl;
    int nbadsyms;
    char *lastname = 0;

    nl = nls = (struct nlist *) malloc((nnames + 1) * sizeof(*nl));
    for (v = vars; v; v = v->next)
	for (inamep = v->inames; *inamep; ++inamep)
	    (nl++)->n_name = *inamep;
    nl->n_name = "";
    if (kname) {
	if ((nbadsyms = nlist(kname, nls)) == -1)
	    goto badkern;
    }
    else {
	if ((nbadsyms = nlist("/vmunix", nls)) == -1
	    && (nbadsyms = nlist("/unix", nls)) == -1
	    && (nbadsyms = nlist("/bsd", nls)) == -1
	    && (nbadsyms = nlist("/hp-ux", nls)) == -1)
	    goto badkern;
    }
/*    printf("couldn't get %d symbols\n", nbadsyms);*/
    nl = nls;
    for (v = vars; v; v = v->next) {
	for (inamep = v->inames; *inamep; ++inamep) {
#if 0
	    printf("setting %s (%s, value = 0x%x)\n", v->inames[0],
		   nl->n_name, *(int *)v->value);
#endif
	    if (nl->n_type != 0L) {
		lseek(kfd, nl->n_value, 0);
		if (write(kfd, v->value, v->len) < v->len)
		    fprintf(stderr, "%s: couldn't write %s into kernel.\n",
			    cmdname, nl->n_name);
		else
		    v->foundname = *inamep;
	    }
	    lastname = nl->n_name;
	    ++nl;
	}
	if (!v->foundname && existwarn)
	    fprintf(stderr, "%s: warning: couldn't find %s to write\n",
		    cmdname, v->inames[0]);
	    
    }
    free(nls);
    return;
  badkern:
    perror("netconfig: nlist");
    fprintf(stderr, "%s: could not get symbols from kernel\n", cmdname);
}
